#!/usr/bin/env node

const chalk = require('chalk');
const ora = require('ora');
const program = require('commander');
const shell = require('shelljs');
const inquirer = require('inquirer');
const spinner = ora('Installing ');
const spawn = require('child_process').spawn;
const mysql = require('mysql');
const fs = require('fs');
const untildify = require('untildify');
//const fs = require('fs-extra')
var path = require("path");
const os = require('os');
//https://github.com/jprichardson/node-fs-extra/blob/master/docs/pathExists.md
//let apt = ['install', '-y', 'libssl-dev', 'libminizip-dev', 'libmysqlclient-dev', "cmysql", 'uuid-dev', ];

let apt = ['install', '-y', 'software-properties-common', 'python-software-properties', 'libminizip-dev'];
let githubRepo = 'https://github.com/brokenhandsio/SteamPressExample.git';

if (!shell.which('wget')) {
  apt.push('wget');
}

if (!shell.which('git')) {
  apt.push('git');
}

if (!shell.which('mysql')) {
  console.error('MySQL not found. Please install MySQL first.');
  spinner.fail('Installation failed');
  shell.exit(1);
}

var installationPathQuestion = [{
  type: 'list',
  name: 'installationPath',
  message: 'Choose app location:',
  choices: ['current working directory', 'custom path']
}];

var mySQLQuestions = [{
  type: 'input',
  name: 'hostname',
  message: 'Database host:',
  default: 'localhost'
},
{
  type: 'input',
  name: 'port',
  message: 'Port:',
  default: '3306'
},
{
  type: 'input',
  name: 'user',
  message: 'Username:',
  validate: function (value) {
    if (value == '') {
      return 'Please enter a username.';
    } else {
      return true;
    }
  }
},
{
  type: 'password',
  name: 'password',
  message: 'Password:',
  validate: function (value) {
    if (value == '') {
      return 'Please enter a password.';
    } else {
      return true;
    }
  }
},
{
  type: 'input',
  name: 'database',
  message: 'Database name:',
  validate: function (value) {
    if (value == '') {
      return 'Please enter a database name.';
    } else {
      return true;
    }
  }
}
];

var mySQLConnectionErrorQuestion = [{
  type: 'confirm',
  name: 'mySQLConnectionRetry',
  message: chalk.red('Could not connect to database. ') + 'Try again?'
}];
var customInstallationPathQuestion = [{
  type: 'input',
  name: 'customInstallationPath',
  message: 'Enter custom path:',
  validate: function (value) {
    if (value == '') {
      return 'Please enter a valid path.';
    } else {
      return true;
    }
  }
}];

var installNginxQuestion = [{
  type: 'confirm',
  name: 'nginx',
  message: 'Nginx not found. Install it now?',
  default: true
}];

var configureNginxQuestion = [{
  type: 'confirm',
  name: 'configure',
  message: 'Do you want to configure nginx now?',
  default: true
}];

siteSettingsQuestions = [{
  type: 'input',
  name: 'hostname',
  message: 'Hostname (or domain):',
  default: 'localhost'
},
{
  type: 'confirm',
  name: 'ssl',
  message: 'Use SSL/TLS?',
  default: false
}
];

siteSSLQuestions = [{
  type: 'input',
  name: 'certificate',
  message: 'SSL certificate path:'
},
{
  type: 'input',
  name: 'key',
  message: 'SSL private key path:'
},
{
  type: 'input',
  name: 'dhparam',
  message: 'Dhparam.pem file path:',
  default: '/etc/nginx/ssl/dhparam.pem'
}];

program
  .option('-d, --debug', 'builds with debug symbols.')
  .option('-v, --verbose', 'print build logs.')
  .parse(process.argv);


//console.log(". = %s", path.resolve(".")); // /root/Documents/test
//console.log("__dirname = %s", path.resolve(__dirname)); // /usr/local/lib/node_modules/demo-cms-cli
//console.log("resolved = %s", path.resolve("999")); // 


//DEPENDENCIES
/*
if (shell.exec('wget -q https://repo.vapor.codes/apt/keyring.gpg -O- | sudo apt-key add -').code !== 0) {
  //shell.echo('Error: Add vapor apt failed');
  spinner.fail('Installation failed');
  shell.exit(1);
}

if (shell.exec('echo "deb https://repo.vapor.codes/apt $(lsb_release -sc) main" | sudo tee /etc/apt/sources.list.d/vapor.list').code !== 0) {
  spinner.fail('Installation failed');
  shell.exit(1);
}

shell.exec('sudo apt-get update', {
    stdio: 'inherit'
  });

if (!shell.which('swift', {
    stdio: 'inherit'
  }) {
  shell.exec('sudo apt-get install swift');  
}

if (!shell.which('vapor')) {
  shell.exec('sudo apt-get install vapor');  
}
*/

//TODO - add npx
/*
var command = 'sudo apt';
//spinner.start();
let child = spawn(command, args, { stdio: 'inherit', shell: true });
child.on('error', (err) => {
  console.error('Failed to start build process.'); //то есть это выполнится если не удастся запустить child process  
});

child.on('exit', (code) => {
  if (code == 0) {
    spinner.succeed('Build succeeded');
  } else {
    spinner.fail('Build failed');
  }
});

child.on('close', (code) => {
  process.exit(code);
});
*/

console.log(chalk.cyan('Installing Demo...'));
inquirer.prompt(installationPathQuestion).then(answers => {
  if (answers.installationPath == 'current working directory') {
    //console.log(path.resolve(process.cwd()));
    installTo(path.resolve(process.cwd()));
  } else {
    inquirer.prompt(customInstallationPathQuestion).then(answers => {
      var resolvedPath = path.resolve(untildify(answers.customInstallationPath));
      //console.log(resolvedPath);      
      installTo(resolvedPath);
    });
  }
});

//TODO: вопрос - создавать ли отдельную папку (как при клонировании, или нет)? Сейчас папка не создается
//еще вариант - при выборе "current working directory создавать папку, а при выборе пути - нет (так как при выборе пути к несуществующей папке она автоматически создается)

function cloneTo(path) {
  //let child = spawn('git', ['-C', path, 'clone', 'https://github.com/brokenhandsio/SteamPressExample.git'],  { stdio: 'inherit', shell: true });
  let child = spawn('git', ['clone', githubRepo, path], {
    stdio: 'inherit',
    shell: true
  });
  child.on('exit', (code) => {
    if (code == 0) {
      spinner.succeed('Clone succeeded');
      build(path);
    } else {
      spinner.fail('Installation failed');
      process.exit(code);
    }
  });
}

//build
function build(path) {
  //console.log("Build path is: " + path);
  let child = spawn('swift', ['build', '--package-path', path, '-c', 'release'], {
    stdio: 'inherit',
    shell: true
  });
  child.on('exit', (code) => {
    if (code == 0) {
      spinner.succeed('Build succeeded');
      setUpMySQL(path);
    } else {
      spinner.fail('Build failed');
      process.exit(code);
    }
  });
}

function installTo(path) {
  cloneTo(path);
}


var setUpMySQL = function (path) {
  console.log(chalk.cyan('Setting up MySQL database...'));
  inquirer.prompt(mySQLQuestions).then(answers => {
    var conn = mysql.createConnection({
      host: answers.hostname,
      port: answers.port,
      user: answers.user,
      password: answers.password,
      database: answers.database
    });
    conn.connect(function (err) {
      if (err) {
        inquirer.prompt(mySQLConnectionErrorQuestion).then(answers => {
          if (!answers.mySQLConnectionRetry) {
            spinner.fail('Installation failed');
            process.exit(1);
          } else {
            setUpMySQL(path);
          }
        });
      } else {
        spinner.succeed('Connected to database')
        var data = JSON.stringify(answers, null, ' ');
        //var configDir = path.replace(/^(.+?)\/*?$/, "$1") + '/Config/'
        var configDir = path + '/Config/'
        //if (workDir.indexOf('~') > -1) {
        //  workDir = untildify(workDir);
        // }
        fs.writeFile(configDir + 'mysql.json', data, (err) => {
          if (err) {
            console.error(err);
            spinner.fail('Installation failed');
            process.exit(1);
            //throw err;
          } else {
            spinner.succeed('Config updated');
            if (!shell.which('nginx')) {
              inquirer.prompt(installNginxQuestion).then(answers => {
                if (answers.nginx) {
                  installNginx();
                  configureNginx(path);
                } else {
                  finish();                  
                }
              });
            } else {
              inquirer.prompt(configureNginxQuestion).then(answers => {
                if(!answers.configure) {
                  finish();
                } else {
                  configureNginx(path);
                }                
              });
              //do you want to configure nginx now?
              
            }
          }
        });
      }
    });
  });
}

var installNginx = function () {
  console.log(chalk.cyan('Installing Nginx...'));
  if (shell.exec('sudo apt-get install nginx', {
    stdio: 'inherit'
  }).code !== 0) {
    spinner.fail('Installation failed');
    shell.exit(1);
  }
}
//TODO: как вариант для оптимизации - читать файл конфига, на лету менять переменные и записывать заново, а не страдать фигней через sed
var configureNginx = function (installPath) {  
  console.log(chalk.cyan('Configure Nginx...'));
  inquirer.prompt(siteSettingsQuestions).then(answers => {
    var hostname = answers.hostname;
    if (!answers.ssl) {
      configureWithoutSSL(hostname, installPath);      
    } else {
      configureSSL(hostname, installPath);
    }    
  });
}
//TODO: как вариант для оптимизации - читать файл конфига, на лету менять переменные и записывать заново, а не страдать фигней через sed
var configureSSL = function (hostname, installPath) {
  console.log(chalk.cyan('Configure SSL/TLS...'));
  inquirer.prompt(siteSSLQuestions).then(answers => {
    var defaultConfPath = path.resolve(__dirname) + '/conf/nginx/kami-ssl.conf';
    var certificatePath = path.resolve(answers.certificate);
    var keyPath = path.resolve(answers.key);
    var dhparamPath = path.resolve(answers.dhparam);

    //copy
    if (shell.exec(`sudo cp ${defaultConfPath} /etc/nginx/sites-available/${hostname}`, {
      stdio: 'inherit'
    }).code !== 0) {
      spinner.fail('Failed to create nginx configuration file.');
      shell.exit(1);
    }

    //replace  
    var serverName = `sudo sed -i 's|{NAME}|${hostname}|g' /etc/nginx/sites-available/${hostname}`;
    if (os.platform() == 'darwin') {
      serverName = `sudo sed -i '' 's|{NAME}|${hostname}|g' /etc/nginx/sites-available/${hostname}`;
    }

    var serverPath = `sudo sed -i 's|{PATH}|${installPath}|g' /etc/nginx/sites-available/${hostname}`;
    if (os.platform() == 'darwin') {
      serverPath = `sudo sed -i '' 's|{PATH}|${installPath}|g' /etc/nginx/sites-available/${hostname}`;
    }

    var certificate = `sudo sed -i 's|{CERTIFICATE}|${certificatePath}|g' /etc/nginx/sites-available/${hostname}`;
    if (os.platform() == 'darwin') {
      certificate = `sudo sed -i '' 's|{CERTIFICATE}|${certificatePath}|g' /etc/nginx/sites-available/${hostname}`;
    }

    var key = `sudo sed -i 's|{KEY}|${keyPath}|g' /etc/nginx/sites-available/${hostname}`;
    if (os.platform() == 'darwin') {
      key = `sudo sed -i '' 's|{KEY}|${keyPath}|g' /etc/nginx/sites-available/${hostname}`;
    }

    var dhparam = `sudo sed -i 's|{DHPARAM}|${dhparamPath}|g' /etc/nginx/sites-available/${hostname}`;
    if (os.platform() == 'darwin') {
      dhparam = `sudo sed -i '' 's|{DHPARAM}|${dhparamPath}|g' /etc/nginx/sites-available/${hostname}`;
    }

    if (shell.exec(serverName, { stdio: 'inherit' }).code !== 0) {
      spinner.fail('Failed to write nginx configuration file.');
      shell.exit(1);
    }
    if (shell.exec(serverPath, { stdio: 'inherit' }).code !== 0) {
      spinner.fail('Failed to write nginx configuration file.');
      shell.exit(1);
    }
    if (shell.exec(certificate, { stdio: 'inherit' }).code !== 0) {
      spinner.fail('Failed to write nginx configuration file.');
      shell.exit(1);
    }
    if (shell.exec(key, { stdio: 'inherit' }).code !== 0) {
      spinner.fail('Failed to write nginx configuration file.');
      shell.exit(1);
    }
    if (shell.exec(dhparam, { stdio: 'inherit' }).code !== 0) {
      spinner.fail('Failed to write nginx configuration file.');
      shell.exit(1);
    }

    //symlink
    createSymlink(hostname);
    reloadNginx();    
  });
}

var configureWithoutSSL = function (hostname, installPath) {  
    var defaultConfPath = path.resolve(__dirname) + '/conf/nginx/kami.conf';
    //copy
    if (shell.exec(`sudo cp ${defaultConfPath} /etc/nginx/sites-available/${hostname}`, {
      stdio: 'inherit'
    }).code !== 0) {
      spinner.fail('Failed to create nginx configuration file.');
      shell.exit(1);
    }
    //replace  
    var serverName = `sudo sed -i 's|{NAME}|${hostname}|g' /etc/nginx/sites-available/${hostname}`;
    if (os.platform() == 'darwin') {
      serverName = `sudo sed -i '' 's|{NAME}|${hostname}|g' /etc/nginx/sites-available/${hostname}`;
    }

    var serverPath = `sudo sed -i 's|{PATH}|${installPath}|g' /etc/nginx/sites-available/${hostname}`;
    if (os.platform() == 'darwin') {
      serverPath = `sudo sed -i '' 's|{PATH}|${installPath}|g' /etc/nginx/sites-available/${hostname}`;
    }

    if (shell.exec(serverName, { stdio: 'inherit' }).code !== 0) {
      spinner.fail('Failed to write nginx configuration file.');
      shell.exit(1);
    }
    if (shell.exec(serverPath, { stdio: 'inherit' }).code !== 0) {
      spinner.fail('Failed to write nginx configuration file.');
      shell.exit(1);
    }
    //symlink
    createSymlink(hostname);
    reloadNginx();
}

var createSymlink = function (path) {
  if (shell.exec(`sudo ln -sf /etc/nginx/sites-available/${path} /etc/nginx/sites-enabled/`, {
    stdio: 'inherit'
  }).code !== 0) {
    spinner.fail('Configuration failed. You can configure nginx manually later.');
    shell.exit(1);
  }
}

var reloadNginx = function () {
  var reload = 'sudo systemctl reload nginx'
  if (os.platform() == 'darwin') {
    reload = `nginx -s reload`
  }
  if (shell.exec(reload, {
    stdio: 'inherit'
  }).code !== 0) {
    spinner.fail('Configuration failed. You can configure nginx manually later.');
    shell.exit(1);
  } else {
    spinner.succeed('Nginx config updated');
    finish();
  }
}

var finish = function () {
spinner.succeed("Done! Demo has been installed successfully. Now you can launch the application using command " + chalk.greenBright("'demo run'") + ' in your working directory. For more information, please visit our Help Center ' + chalk.greenBright('http://demo.com/docs/'));
shell.exit(0);  
}